use board_shared::{
    board::Board,
    position::{Grid2d, Position2d},
    tile::{Digit, Operator, Tile},
};
use serde::{Deserialize, Serialize};

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct BoardRef {
    pub tiles: Vec<Option<TileRef>>,
    pub width: usize,
    pub height: usize,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub enum TileRef {
    Digit(DigitRef),
    Operator(OperatorRef),
    Equals,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct DigitRef {
    pub value: i8,
    pub has_left_parenthesis: bool,
    pub has_right_parenthesis: bool,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub enum OperatorRef {
    Add,
    Subtract,
    Multiply,
    Divide,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Position2dRef {
    pub x: usize,
    pub y: usize,
}

impl From<&Board> for BoardRef {
    fn from(value: &Board) -> Self {
        Self {
            tiles: value
                .iter()
                .map(|tile| tile.map(Into::into))
                .collect::<Vec<_>>(),
            width: value.width(),
            height: value.height(),
        }
    }
}

impl From<Tile> for TileRef {
    fn from(value: Tile) -> Self {
        match value {
            Tile::Digit(digit) => TileRef::Digit(digit.into()),
            Tile::Operator(operator) => TileRef::Operator(operator.into()),
            Tile::Equals => TileRef::Equals,
        }
    }
}

impl From<Digit> for DigitRef {
    fn from(value: Digit) -> Self {
        Self {
            value: value.value,
            has_left_parenthesis: value.has_left_parenthesis,
            has_right_parenthesis: value.has_right_parenthesis,
        }
    }
}

impl From<Operator> for OperatorRef {
    fn from(value: Operator) -> Self {
        match value {
            Operator::Add => OperatorRef::Add,
            Operator::Subtract => OperatorRef::Subtract,
            Operator::Multiply => OperatorRef::Multiply,
            Operator::Divide => OperatorRef::Divide,
        }
    }
}

impl From<Position2d> for Position2dRef {
    fn from(value: Position2d) -> Self {
        Self {
            x: value.x,
            y: value.y,
        }
    }
}

impl From<Position2dRef> for Position2d {
    fn from(value: Position2dRef) -> Self {
        Self {
            x: value.x,
            y: value.y,
        }
    }
}
